import socket
import threading
import random
import time

class GSMSimulator:
    def __init__(self, host='0.0.0.0', port=5555):
        self.server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server_socket.bind((host, port))
        self.server_socket.listen(2)
        self.clients = {}
        print("GSM Simulator started. Waiting for connections...")

    def simulate_signal_strength(self):
        return random.uniform(0, 1)

    def handle_client(self, client_socket, address):
        try:
            client_socket.settimeout(5)
            role = client_socket.recv(1024).decode().strip()
            if role not in ["CALLER", "RECEIVER"]:
                print(f"Invalid role received from {address}: {role}")
                client_socket.close()
                return
            self.clients[role] = client_socket
            print(f"{role} connected from {address}")

            if role == "CALLER":
                self.handle_caller(client_socket)
            elif role == "RECEIVER":
                self.handle_receiver(client_socket)
        except Exception as e:
            print(f"Error handling client {address}: {e}")
        finally:
            if role in self.clients:
                del self.clients[role]
            client_socket.close()

    def handle_caller(self, caller_socket):
        print("Caller connected, waiting for receiver...")
        while "RECEIVER" not in self.clients:
            time.sleep(1)
        receiver_socket = self.clients["RECEIVER"]
        print("Receiver found, sending RINGING...")
        receiver_socket.send("RINGING".encode())

        while True:
            try:
                data = caller_socket.recv(1024).decode().strip()
                if not data:
                    print("Caller disconnected unexpectedly.")
                    receiver_socket.send("CALL_END".encode())
                    break
                if data == "CALL_END":
                    print("Call terminated by caller.")
                    receiver_socket.send("CALL_END".encode())
                    break

                signal = self.simulate_signal_strength()
                if signal < 0.2:
                    print("Call dropped due to low signal strength.")
                    caller_socket.send("CALL_DROPPED".encode())
                    receiver_socket.send("CALL_DROPPED".encode())
                    break

                print(f"Relaying to receiver: {data} (Signal: {signal:.2f})")
                receiver_socket.send(data.encode())
                time.sleep(0.1)  # Small delay to ensure data is sent properly
            except Exception as e:
                print(f"Error in caller loop: {e}")
                receiver_socket.send("CALL_END".encode())
                break

    def handle_receiver(self, receiver_socket):
        print("Receiver connected, waiting for data...")
        try:
            while True:
                data = receiver_socket.recv(1024)
                if not data:
                    print("Receiver disconnected.")
                    break
        except Exception as e:
            print(f"Error in receiver loop: {e}")

    def run(self):
        while True:
            client_socket, address = self.server_socket.accept()
            print(f"New connection from {address}")
            thread = threading.Thread(target=self.handle_client, args=(client_socket, address))
            thread.start()

if __name__ == "__main__":
    gsm = GSMSimulator()
    gsm.run()