import os
import struct
import time
import zlib
import hashlib

def crc32_of(data: bytes) -> int:
    """
    Compute CRC-32 of 'data'.
    """
    return zlib.crc32(data) & 0xffffffff


# ---------------------------------------------------------------------------
# PING REQUEST (new format)
# Fields (in order):
#   - session_nonce: 129 bits (from the top 129 bits of 17 random bytes)
#   - version: 7 bits
#   - cipher: 4 bits   (0 = AES-256-GCM, 1 = ChaCha20-poly1305; for now only 0 is used)
#   - CRC: 32 bits
#
# Total bits: 129 + 7 + 4 + 32 = 172 bits. We pack into 22 bytes (176 bits) with 4 spare bits.
# ---------------------------------------------------------------------------
def build_ping_request(version: int, cipher: int, nonce_full: bytes = None) -> bytes:
    """
    Build a Ping request with:
      - session_nonce: 129 bits (derived from 17 random bytes by discarding the lowest 7 bits)
      - version: 7 bits
      - cipher: 4 bits   (0 = AES-256-GCM, 1 = ChaCha20-poly1305; we use 0 for now)
      - CRC: 32 bits

    Total = 129 + 7 + 4 + 32 = 172 bits.
    We pack into 22 bytes (176 bits) leaving 4 unused bits.
    """
    if not (0 <= version < 128):
        raise ValueError("Version must fit in 7 bits (0..127)")
    if not (0 <= cipher < 16):
        raise ValueError("Cipher must fit in 4 bits (0..15)")
    # Generate 17 random bytes if none provided (17 bytes = 136 bits)
    if nonce_full is None:
        nonce_full = os.urandom(17)
    if len(nonce_full) < 17:
        raise ValueError("nonce_full must be at least 17 bytes")
    # Use the top 129 bits of the 136 bits:
    nonce_int_full = int.from_bytes(nonce_full, 'big')  # 136 bits
    nonce_129_int = nonce_int_full >> 7  # drop the lowest 7 bits => 129 bits
    # Convert the derived 129-bit nonce to 17 bytes.
    # Since 129 bits < 17*8=136 bits, the top 7 bits of the result will be 0.
    nonce_129_bytes = nonce_129_int.to_bytes(17, 'big')

    # Pack the fields: shift the 129-bit nonce left by (7+4)=11 bits, then add version (7 bits) and cipher (4 bits).
    partial_int = (nonce_129_int << 11) | (version << 4) | (cipher & 0x0F)
    # This partial data is 129+7+4 = 140 bits; we pack into 18 bytes (144 bits) with 4 spare bits.
    partial_bytes = partial_int.to_bytes(18, 'big')
    # Compute CRC over these 18 bytes.
    cval = crc32_of(partial_bytes)
    # Combine the partial data with the 32-bit CRC.
    final_int = (int.from_bytes(partial_bytes, 'big') << 32) | cval  # 140 + 32 = 172 bits
    final_bytes = final_int.to_bytes(22, 'big')
    # Optionally, store or print nonce_129_bytes (the session nonce) rather than the original nonce_full.
    return final_bytes

def parse_ping_request(data: bytes):
    """
    Parse a Ping request (22 bytes = 172 bits).
    Returns (session_nonce_bytes, version, cipher) or None if invalid.
    The session_nonce_bytes will be 17 bytes, representing the 129-bit value.
    """
    if len(data) != 22:
        return None
    final_int = int.from_bytes(data, 'big')  # 176 bits integer; lower 32 bits = CRC, higher 140 bits = partial
    crc_in = final_int & 0xffffffff
    partial_int = final_int >> 32  # 140 bits
    partial_bytes = partial_int.to_bytes(18, 'big')
    crc_calc = crc32_of(partial_bytes)
    if crc_calc != crc_in:
        return None
    # Extract fields: lowest 4 bits: cipher; next 7 bits: version; remaining 129 bits: session_nonce.
    cipher = partial_int & 0x0F
    version = (partial_int >> 4) & 0x7F
    nonce_129_int = partial_int >> 11  # 140 - 11 = 129 bits
    # Convert to 17 bytes. Since the number is < 2^129, the top 7 bits will be zero.
    session_nonce_bytes = nonce_129_int.to_bytes(17, 'big')
    return (session_nonce_bytes, version, cipher)



# ---------------------------------------------------------------------------
# PING RESPONSE (new format)
# Fields:
#   - timestamp: 32 bits (we take the lower 32 bits of the time in ms)
#   - version: 7 bits
#   - cipher: 4 bits
#   - answer: 1 bit
#   - CRC: 32 bits
#
# Total bits: 32 + 7 + 4 + 1 + 32 = 76 bits; pack into 10 bytes (80 bits) with 4 spare bits.
# ---------------------------------------------------------------------------
def build_ping_response(version: int, cipher: int, answer: int) -> bytes:
    if not (0 <= version < 128):
        raise ValueError("Version must fit in 7 bits")
    if not (0 <= cipher < 16):
        raise ValueError("Cipher must fit in 4 bits")
    if answer not in (0, 1):
        raise ValueError("Answer must be 0 or 1")
    t_ms = int(time.time() * 1000) & 0xffffffff  # 32 bits
    # Pack timestamp (32 bits), then version (7 bits), cipher (4 bits), answer (1 bit): total 32+7+4+1 = 44 bits.
    partial_val = (t_ms << (7+4+1)) | (version << (4+1)) | (cipher << 1) | answer
    partial_bytes = partial_val.to_bytes(6, 'big')  # 6 bytes = 48 bits, 4 spare bits.
    cval = crc32_of(partial_bytes)
    final_val = (int.from_bytes(partial_bytes, 'big') << 32) | cval  # 44+32 = 76 bits.
    final_bytes = final_val.to_bytes(10, 'big')  # 10 bytes = 80 bits.
    return final_bytes

def parse_ping_response(data: bytes):
    if len(data) != 10:
        return None
    final_int = int.from_bytes(data, 'big')  # 80 bits
    crc_in = final_int & 0xffffffff
    partial_int = final_int >> 32  # 48 bits
    partial_bytes = partial_int.to_bytes(6, 'big')
    crc_calc = crc32_of(partial_bytes)
    if crc_calc != crc_in:
        return None
    # Extract fields: partial_int has 48 bits. We only used 44 bits for the fields.
    # Discard the lower 4 spare bits.
    partial_int >>= 4  # now 44 bits.
    # Now fields: timestamp: 32 bits, version: 7 bits, cipher: 4 bits, answer: 1 bit.
    answer = partial_int & 0x01
    cipher = (partial_int >> 1) & 0x0F
    version = (partial_int >> (1+4)) & 0x7F
    timestamp = partial_int >> (1+4+7)
    return (timestamp, version, cipher, answer)


# =============================================================================
# 3) Handshake
#    - 32-bit timestamp
#    - 64-byte ephemeral pubkey (raw x||y = 512 bits)
#    - 64-byte ephemeral signature (raw r||s = 512 bits)
#    - 32-byte PFS hash (256 bits)
#    - 32-bit CRC
#    => total 4 + 64 + 64 + 32 + 4 = 168 bytes = 1344 bits
# =============================================================================

def build_handshake_message(timestamp: int,
                            ephemeral_pubkey: bytes,
                            ephemeral_signature: bytes,
                            pfs_hash: bytes) -> bytes:
    """
    Build handshake:
      - 4 bytes: timestamp
      - 64 bytes: ephemeral_pubkey (x||y, raw)
      - 64 bytes: ephemeral_signature (r||s, raw)
      - 32 bytes: pfs_hash
      - 4 bytes: CRC-32
    => 168 bytes total
    """
    if len(ephemeral_pubkey) != 64:
        raise ValueError("ephemeral_pubkey must be 64 bytes (raw x||y).")
    if len(ephemeral_signature) != 64:
        raise ValueError("ephemeral_signature must be 64 bytes (raw r||s).")
    if len(pfs_hash) != 32:
        raise ValueError("pfs_hash must be 32 bytes.")

    partial = struct.pack("!I", timestamp) \
              + ephemeral_pubkey \
              + ephemeral_signature \
              + pfs_hash
    cval = crc32_of(partial)
    return partial + struct.pack("!I", cval)


def parse_handshake_message(data: bytes):
    """
    Parse handshake message (168 bytes).
    Return (timestamp, ephemeral_pub, ephemeral_sig, pfs_hash) or None if invalid.
    """
    if len(data) != 168:
        return None
    partial = data[:-4]   # first 164 bytes
    crc_in = struct.unpack("!I", data[-4:])[0]
    crc_calc = crc32_of(partial)
    if crc_calc != crc_in:
        return None

    # Now parse fields
    timestamp = struct.unpack("!I", partial[:4])[0]
    ephemeral_pub = partial[4:4+64]
    ephemeral_sig = partial[68:68+64]
    pfs_hash = partial[132:132+32]
    return (timestamp, ephemeral_pub, ephemeral_sig, pfs_hash)


# =============================================================================
# 4) PFS Hash Helper
#    If no previous session, return 32 zero bytes
#    Otherwise, compute sha256(session_number || last_shared_secret).
# =============================================================================

def compute_pfs_hash(session_number: int, shared_secret_hex: str) -> bytes:
    """
    Return 32 bytes (256 bits) for the PFS field.
    If session_number < 0 => means no previous session => 32 zero bytes.
    Otherwise => sha256( session_number (4 bytes) || shared_secret ).
    """
    if session_number < 0:
        return b"\x00" * 32

    # Convert shared_secret_hex to raw bytes
    secret_bytes = bytes.fromhex(shared_secret_hex)
    # Pack session_number as 4 bytes
    sn_bytes = struct.pack("!I", session_number)
    return hashlib.sha256(sn_bytes + secret_bytes).digest()
